//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_SHAPE_INCLUDED
#include "svg_element.h"
#define SVG_SHAPE_INCLUDED

namespace svg
{
//////////////////////////////////////////////////////////////////////
// shape: an element of the SVG graph, which can be drawn on the
//		  canvas (i.e. final picture) - base class for all graphic 
//		  objects, like text, path, circle, poly, rectangle.
//		  It is also the base class for the group of them, and 
//		  their references (use, tref)
//////////////////////////////////////////////////////////////////////
class shape : public element 
{
	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		shape(element *parent, unsigned level): element(parent, level),
		m_bounds(0,0,-1,-1), //initial value: invalid bounds
		m_filter(invalid_handle),
		m_mask(invalid_handle),
		m_clip_path(invalid_handle),
		m_opacity(invalid_handle)
		{ m_display = true; }

		virtual void on_attribute(parser &prs, const char * name, const char *value);
		
		//helper method: renders the shape. 
		//Arguments:
		//	req:	request information
		//Method used within on_request, also by the filters to
		//render the shape on a buffer, specified in the req content.
		//Returns id of the last rendered child element, incremented by 1.
		//(Typical result, returned by on_request)
		virtual handle render(request &req) = 0;
		
		virtual bool is_of(const char *cls) 
		{
			if(strcmp("shape",cls) == 0)
			{
				return true;
			}
			else //it is not your name? Make sure, asking your parent :)
				return element::is_of(cls);
		}

	public:

		//////////////////////////////////////////////////////////////////////
		// New virtual methods
		//////////////////////////////////////////////////////////////////////
		//request handler:
		virtual handle on_request(request &req); 

		//////////////////////////////////////////////////////////////////////
		// Other methods and fields
		//////////////////////////////////////////////////////////////////////
		//properties to override (by the tex:filter, tex:mask, tex:clip_path,
		//elements, for example)
		bool display() {return m_display;}
		void display(bool on) {m_display = on;}

		//handle filter() {return m_filter;}
		//void filter(handle ref) {m_filter = ref;}

		//handle mask() {return m_mask;}
		//void mask(handle ref) {m_mask = ref;}
		
		//handle clip_path() {return m_clip_path;}
		//void clip_path(handle ref) {m_clip_path = ref;}

		//handle opacity() {return m_opacity;}
		//void opacity(handle ref) {m_opacity = ref;}
		
		//Returns actual style settings:
		template<class StyleRequest> handle get_style(StyleRequest &req)
		{
			req.style.update(req,m_attribute_first, m_attribute_last);
			return skip(req); //this is the "to_root" request - no forwarding!
		}
		
		//Returns element's transformation: 
		//(needed in the svg::filter)
		transform trans() {return m_trans;}

		//Returns actual bounding box (required for gradients or paterns
		//with units set to objectBoundingBox). Bounding box describes the
		//extents of the object in its "natural" coordinates, before any
		//transformation has been applied.
		//Sets the internal m_bounds field, on first use.
		rectangle bounding_box();

		//Returns extents of a transformed shape
		//Arguments:
		//	trans:			transformation, which should be applied before
		//					the measurement
		//	force_measure:	forces to measure (used in bounding_box(), to avoid
		//					endless recursion)
		rectangle shape::bounding_box(transform trans, bool force_measure = false);

		//Some details of the style or referenced objects have to be prepared
		//before each rendering. This procedure does it.
		//Arguments
		//	req:	render request (just to retrieve elements or attributes)
		//	style:	the styleset, that will be used in this rendering
		//Used in the render() procedure of various child classes (svg::path, 
		//svg::tspan, ...)
		void prepare_render(request &req, styleset &style);

	private:

		//Helper procedures to handle the render request:
		handle render_with_filter(request &req);
		handle render_with_mask(request &req);
		handle render_with_clip_path(request &req);
		handle render_with_opacity(request &req);

	protected:
		//////////////////////////////////////////////////////////////////////
		// Instance additional fields
		//////////////////////////////////////////////////////////////////////

		transform	m_trans;
		rectangle	m_bounds;  //local box of this shape
							   //set only when is needed (a gradient or a pattern
							   //with units set to objectBoundingBox is applied)
							   //otherwise - m_bounde.is_valid() = false.
		bool		m_display; //true, when shape is to be rendered

		//All three: clip_path, mask, filter are in fact three 
		//kinds of a raster filter:
		handle		m_filter;    //id of the eventual filter:	
		handle		m_mask;      //id of the eventual mask:	
		handle		m_clip_path; //id of the eventual clip path, used to restrict
								 //this shape:	
		handle		m_opacity;	 //id of the eventual master opacity filter
								 //(for items partially transparent)

}; //element

}; //namespace svg
#endif //SVG_SHAPE_INCLUDED

