//----------------------------------------------------------------------------
// svgtex - a SVG vector texture plugin for Blender
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
//
// This program is based on:	
//			vectex - a SVG vector texture plugin for Blender
//			(Copyright (C) 2008 Marcelo de Gomensoro Malheiros, 
//			 mgmalheiros@gmail.com)
// 
// Contains elements from:
//			AGG - a graphic library 
//			(Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
//
// Uses functions from:
//          Expat - a XML parser library 
//			(Copyright James Clark)
// 
// svgtext is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#ifndef TEX_INSTANCE_INCLUDED
#include "svgtex.h"
#include "svg_root.h"
#include "tex_tileset.h"
#define TEX_INSTANCE_INCLUDED
namespace tex
{
//////////////////////////////////////////////////////////////////////
// instance: the data, based on the Cast structure, received from Blender 	
//			 Contains a set of texture images (tex::tileset), and a svg 
//			 graph (svg::root), to render new images, when needed. 
//////////////////////////////////////////////////////////////////////

class instance
{
public:

	enum extend_mode
	{	//synchronize these constants with stnames array from svgtex.cpp:
		extent_extend	= 0,
		extent_repeat	= 1,
		extent_clip		= 2
	};

	enum constants
	{
	};

protected: //data fields:
	bool			m_enabled;	  //flag, set after successfull initialization
	tileset			m_tileset;    //the set of texture images
	svg::root		m_svg_graph;  //image renderer
	color			m_background; //background color
	int				m_default_LOD;//default level of details 
								  //(default resolution = 2^m_default_LOD)
								  //used when Blender's OSA (rendering oversampling)
								  //is disabled.
	unsigned char	m_file_name[file_name_size];//Source file name
public:
	//Constructor
	//Arguments:
	//	stype:	number of extend mode
	//	cast:	the data structure with parameters, received from Blender
	instance(Cast *cast);

	virtual ~instance()
	{
		MSG0("- instance destroyed\n");
	}
	//Static constructor: creates a new instance, updates the cast structure
	//Arguments:
	//	cast:	the data structure with parameters, received from Blender
	static instance *create_for(Cast *cast)
	{
		cast->mem = 0;
		cast->instance_data =  new instance(cast);
		return (instance *) (cast->instance_data);
	}

	//Static destructor: removes the instance, updates the cast structure
	//Arguments:
	//	cast:	the data structure with parameters, received from Blender
	static void discard_from(Cast *cast)
	{
		if (cast->instance_data) 
		{
			instance *data = (instance *) cast->instance_data;
			delete data;
			cast->instance_data = NULL;
			cast->mem = 0;
		}
	}

	// --- some helper functions - protected field acessors -----------------
	//Returns the resolution, used by default when OSA is disabled
	inline int32u default_resolution() {return 1 << m_default_LOD;}

	inline bool is_enabled() {return m_enabled;}

	inline tileset &tiles() {return m_tileset;}
	
	inline const char *file_name() {return (const char *) m_file_name+0;}
	
	inline color &background() {return m_background;}
	void background(color &c) {m_background = c;}

	inline int default_LOD() {return m_default_LOD - tile_level;}
	inline void default_LOD(int LOD) {m_default_LOD = LOD + tile_level;}

	// --- other methods: ---------------------------------------------------
	//Renders a tileset image (representd as <pixf>)
	//Arguments:
	//	level:	level of details of the image (total resolution = 2^level)
	//	row:	vertical position of the tile on the picture;
	//	col:	horizontal position of the tile on the picture;
	void render(int32 level, int32u col, int32u row, pixfmt &pixf);
	
	//Returns a texel for given texture coordinates
	//Arguments:
	//	mode:	 selected mode of texture extension	(expand, clip, repeat)
	//	u,v:	 the texture coordinates, received from Blender (negaitive or 
	//			 positive, range -1.0 .. 1.0 matches the picture size)
	//	dxt,dyt: optional: two vectors (dxt[0],dyt[0]), (dxt[1],dyt[1]) 
	//			 which describe the size of teh pixel. NULL when Blender's OSA 
	//			 is off. In the same texture coordinate space like u,v 
	//	radius:	 radius of the pixel interpolation (0 = no pixel interpolation)
	//	fidelity:a value, describing the level of detail degradation.
	//			 Its range is from -5.0 to 0.0 (0.0 means "no degradation",
	//			 -1.0: classic mipmapping, lower values - "huge mipmapping")	
	texel sample(extend_mode mode, float u, float v, float *dxt, float *dyt, 
											int32u radius, float fidelity = 0.0);

protected:
	//Clamps given texture coord to range (-1.0, 1.0).
	//Arguments:
	//	mode:	 selected mode of texture extension	(expand, clip, repeat)
	//	t:	a texture coordinate, received from Blender
	//		(because all extend_modes are symetric, the same method  
	//		 can be used for both: u and v texture coordinates)
	//Returns false, when the texel for such coordinate should not be 
	//processed, at all
	bool clamp_coord(extend_mode mode,float *t)
	{
		if( *t < -1.0 || *t > 1.0 )
		{
			switch (mode)
			{
			case extent_extend:
			case extent_clip:
								*t = float((*t > 1.0) ?  1.0: -1.0);
								break;
			case extent_repeat:
								*t = (*t + float(1.0))*float(0.5);
								*t = (*t - float(floor(*t)))*float(2.0) - float(1.0);
								break;
			default:
				MSG1("- clamp coord: unknown extend mode = %d\n", mode);
			}//end switch

			return mode != extent_clip;
		}
		else
		{
			return true;
		}
	}
};//end instance

}//end namespace tex
#endif //TEX_INSTANCE_INCLUDED

