//----------------------------------------------------------------------------
// svgtex - a SVG vector texture plugin for Blender
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
//
// This program is based on:	
//			vectex - a SVG vector texture plugin for Blender
//			(Copyright (C) 2008 Marcelo de Gomensoro Malheiros, 
//			 mgmalheiros@gmail.com)
// 
// Contains elements from:
//			AGG - a graphic library 
//			(Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
//
// Uses functions from:
//          Expat - a XML parser library 
//			(Copyright James Clark)
// 
// svgtext is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#include "svgtex.h"
#include "tex_tile.h"
#include "tex_instance.h"

namespace tex
{
	//Initializes this object - second phase of construction!
	//Arguments:
	//	inst:	the instance object - it contains the tileset, which
	//			contains this tile. Needed to render the image.
	//	level:	resolution level (a power of 2 - for example 8 means
	//			an image of 256x256 pixels. Higher levels implies
	//			multi-tile image (where row and col can be > 0)
	//	col:	for single-tile images (up to tile size): 0
	//			for multi-tile images: column of the tile on the
	//			picture (0 is the rightmost one)
	//	row:	for single-tile images (up to tile size): 0
	//			for multi-tile images: row of the tile on the
	//			picture (0 is the topmost one)
	void tile::initialize(instance &inst, int32u level, int32u col, int32u row)
	{
		int resolution = 1 << level; //overall resolution of the image
		//nominal tile size:
		int nom_side = (level < tile_level)? resolution : tile_size; 
		//full tile size:
		int frm_side = tile_border + nom_side + tile_border;
		//Allocate the image memory:
		m_image = create_image(frm_side);
		m_rbuf.attach(m_image, frm_side, frm_side, frm_side * bytes_pp);
		m_pixf.attach(m_rbuf);
		//At this moment m_pixf contains WHOLE buffer, WITH border, while
		//row and col describes the nominal placement
		inst.render(level, col, row, m_pixf);
		if (is_mono())
		{
			memcpy(m_mono, m_image,bytes_pp);
			m_rbuf = rndbuf(); //clear the rendering buffer ...
			m_pixf.attach(m_rbuf); //clear pixel format ...
			m_image = release_image(m_image); //discard image.
			m_image = m_mono;
			m_age = no_aging;
			MSG4(">> tile is of solid color #%02x%02x%02x%02x\n",
										m_mono[2],m_mono[1],m_mono[0],m_mono[3]);
		}
		else
		{
			//from this moment pixf contains the nominal area, only!
			m_pixf.attach(m_pixf, tile_border, tile_border, 
						nom_side + tile_border-1, nom_side + tile_border-1);
			m_age = initial_age; //Just to mark it as the enabled tile. 
								 //The age value will be changed immediately to timestamp, 
								 //in the caller procedure (see tileset::get_tile())
		}
	}

	//Resets tile to the initial (unused) state
	void tile::reset()
	{
		if(is_initialized() && ! is_constant_color())
		{
			m_rbuf = rndbuf();			//clear the rendering buffer ...
			m_pixf.attach(m_rbuf);		//clear pixel format ...
			m_image = release_image(m_image);		//discard image.
			MSG0(" - tile destroyed\n")
		}
		m_age = unused;
	}


}//end namespace tex
