//----------------------------------------------------------------------------
// svgtex - a SVG vector texture plugin for Blender
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
//
// This program is based on:	
//			vectex - a SVG vector texture plugin for Blender
//			(Copyright (C) 2008 Marcelo de Gomensoro Malheiros, 
//			 mgmalheiros@gmail.com)
// 
// Contains elements from:
//			AGG - a graphic library 
//			(Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
//
// Uses functions from:
//          Expat - a XML parser library 
//			(Copyright James Clark)
// 
// svgtext is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#ifndef TEX_TILESET_INCLUDED
#include "tex_tile.h"
#define TEX_TILESET_INCLUDED
namespace tex
{
//////////////////////////////////////////////////////////////////////
// tileset:	a set of the image tiles (tex::tile objects).
//			This set is growing "on demand", and periodically shrinking
//			when its memory limit (set by the user) has been exceeded.
//////////////////////////////////////////////////////////////////////

class tileset
{
public:
	enum constants
	{
		max_age		       = 256000, //when tileset::m_timestamp will reach 
									 //this value - it will be decreased 
									 //(and the age in all tiles)
									 //(but only when the oldest tiles are removed!)
		min_check_interval = 20000, //to avoid often calls to tile management....	

	};

protected: //data fields:


	//Actual state fields:
	int32u	m_max_level;	 //actual highest level of details (power of 2,
							 //which gives the maximum resolution of an
							 //image stored in the tiles. Corresponds to
							 //m_tiles.size(). (With tile_size of 256, you
							 //need a 4 additional tiles (512x512) for 
							 //max_level = 9, another 16 (1024x1024) for 
							 //max_level = 10, another 64 (2048x2048) for
							 //max_level = 11, and so on.

	agg::pod_bvector<tile,9> m_tiles; //Dynamic array of tiles. All tiles for
							 //a stored image are placed in continous index
							 //range of m_tiles array. They are stored "row by row",
							 // - the tile of (col, row) = (0,0) is stored first,
							 //then the tile of (1, 0), then (2,0), and so on.
							 //The second template parameter determines the 
							 //number of tex::tile objects per internal block
							 //(9 = 512, 8 = 256, 7 = 128, ..). This array
							 //grows by allocating another block of memory.
							 //512 tiles is enough to store a mipmap collection,
							 //up to 4096x4096 size.
	
		//m_max_level and m_tileset array grow at user request and never 
		//decreases.They can decrease used memory by releasing some infrequent
		//used images discarding some tiles, instead. See t
	
	int32u	m_timestamp;	 //increases by 1 at every request for a texel
							 //(every call to sample() function, used
							 //once or twice in plugin_tex_doit()).
							 //Its actual value is used to mark the actual
							 //"age" of the tileset, which has been used.
							 //Periodically decreased (in tileset::limit())
	int32u	m_checked_timestamp; //set by tileset::limit() - stores value
								 //of previous call (compared with m_timestamp
								 //for a difference greater than 
								 //tileset::min_check_interval, to avoid
								 //too often calls in helpless situation....)

	//Some statistical fields:
	int32u	m_tile_memory;	 //actual memory, consumed by allocated images
							 //in all enabled tiles.
	int32	m_count_enabled; //actual number of enabled (used) tiles
							 //(counted both: tiles with allocated image
							 //and tiles with constant color)
	int32	m_count_mono;	 //actual number of tiles, which all area
							 //contains a monolithic, constant color

public:
	tileset():
		m_max_level(0),
		m_timestamp(tile::initial_age),
		m_checked_timestamp(tile::initial_age),
		m_tile_memory(0),
		m_count_enabled(0),
		m_count_mono(0)
	{}

	//Resets the tileset to the initial state.
	void reset()
	{
		m_tiles.free_all(); //calls destructor for each tile
		m_max_level = 0;
		m_timestamp = m_checked_timestamp = tile::initial_age;
		m_tile_memory = 0;
		m_count_enabled = 0;
		m_count_mono = 0;
	}

	virtual ~tileset()
	{
		MSG0("- tileset destroyed\n");
	}

	//Returns the memory amount, allocated to this tileset, in megabytes
	int32u used_memory() 
	{	
		return agg::uround(
							double(m_tiles.byte_size() + m_tile_memory)
							/
							double(1 << 20) //1 MB
						  );
	}	

	//return actual maximum Level Of Details:
	inline int32u max_level() {return m_max_level;}

	//increases and returns actual the timestamp
	inline int32u timestamp() {return m_timestamp;}

	//increases and returns actual the timestamp
	inline int32u increase_timestamp() 
	{
		return m_timestamp++;
	}

	//returns the tile, which contains requested pixel. It creates a tile
	//"on the fly", when a new one is required.
	//Arguments:
	//	level:		determines resolution of the picture, which is needed.
	//				(level = 8, => resolution = 256x256,
	//				 level = 9, => resulution = 512x512)
	//	x,y:		pixel coordinates (expressed for full image resolution)
	//	inst:		instance object, required to initialize a new tile.
	tile &get_tile(int32u level, int32 x, int32 y, instance &inst);

	//Arguments:
	//	u,v:		texture coordinates (from -1.0 to 1.0). 
	//				(-1.0, -1.0) corresponds lower left corner of the
	//				picture, (1.0, 1.0) corresponds to upper right corner
	//				of the picture. Texture coordinates passed thera
	//				are always clamped to the range from -1.0 to 1.0;
	//	level:		determines resolution of the picture, which is needed.
	//				(level = 8, => resolution = 256x256,
	//				 level = 9, => resulution = 512x512)
	//	radius:		span of the pixel interpolation (pixels):
	//				0 = no interpolation.
	texel sample(float u, float v, int32u level, int32u radius, instance &inst)
	{
		int32u resolution = 1 << level;
		//tu,tv: pixel coordinates on the image, with subpixel accuracy.
		float tu = (float(u + 1.0) * (resolution / 2)) - float(0.5);
		float tv = (float(1.0 - v) * (resolution / 2)) - float(0.5);

		tile & t = get_tile(level, int32(tu),int32(tv), inst);
		return t.sample(tu,tv,resolution,radius);
	}

	//Verifies the memory, used by this tileset with the requested maximum.
	//When tileset uses more than the given amount - the oldest tiles are
	//reduced.
	//Additionally, it "conserves" the m_timestamp value, reducing it on all 
	//tiles (to not allow to grow it to infinity)
	void limit(int32u max_memory);

protected:
	//Returna index of the first tile (row=0,col=0), that belongs to a picture
	//which have requested level of detail 
	//Arguments:
	//	level:	the power of 2, which determines the picture size 
	//			(level = 8  =>  256x256  - just single tile)
	//			(level = 9  =>  512x512  - first of  4 tiles), 
	//			(level = 10 => 1024x1024 - first of 16 tiles)
	int base_index(int level)
	{
		int index = (level > tile_level) ? tile_level : level;
		int ntiles = 1;
		while (level > tile_level)
		{
			index  += ntiles;
			ntiles *= 4;
			level --;
		}

		return index;
	}

	//timestamp maintenence (to not exceed the int range!)
	void maintain_timestamp();

	//Grows the tile array up to given level
	void grow_to(int32u level)
	{
		if(level > m_max_level)
		{
			int i, new_size = base_index(level+1);
			//add the new tiles
			for (i = m_tiles.size(); i < new_size; i++)
			{
				m_tiles.add(tile()); //each add() call increases the m_tiles.
									 //When an internal memory block is finished
									 //(they have capacity of 512 tiles) - another
									 //one is allocated
			}//end for
			m_max_level = level;
		}
	}

	//Returns index of a first tile, older than given age (i.e. his age() < age)
	//Arguments:
	//	age:	age to compare (in the "timestamp" units)
	//When there is no such tile: returns -1
	int32 oldest(int32 age);

};//end tileset

}//end namespace tex
#endif //TEX_TILESET_INCLUDED

